/*
 * Header file for the udp6 tool
 *
 */


/* Constants for debug mode */
#define PROBE_DUMP			1
#define PROBE_SCRIPT		2

/* Constants for UDP buffers */
#define UDP_BUFFER_SIZE			65535
#define UDP_INPUT_BUFFER_SIZE	TCP_BUFFER_SIZE
#define UDP_OUTPUT_BUFFER_SIZE	TCP_BUFFER_SIZE

